package opusgames.com.sumpairlist;


/**
 * This class takes a sorted list of integers and finds all pairs of elements in the list that sum up 
 * to an input value. Note that this does the processing with 0(n) complexity (a single pass through 
 * the list). This version also allows duplicate values in the list, as long as the list remains sorted.
 * 
 * @author Douglas Lanford
 */
public class SumPairList {

	/** 
	 * find all pairs in a sorted list that sum up to an input value with O(n) complexity
	 * 
	 * @param sortedDataList - sorted list of integers
	 * @param sumValue - sum value
	 * 
	 * output is to the logging console
	 */
	public void findSumPairs(int[] sortedDataList, int sumValue) {
		// since the list is sorted, you can go through the list from both ends at the same time, meeting 
		// in the middle
		int lowIndex = 0;
		int highIndex = sortedDataList.length - 1;

		while (lowIndex < highIndex) {
			while ((lowIndex < highIndex)
					&& (sortedDataList[lowIndex] + sortedDataList[highIndex] < sumValue)) {
				lowIndex++;
			}

			while ((lowIndex < highIndex)
					&& (sortedDataList[lowIndex] + sortedDataList[highIndex] > sumValue)) {
				highIndex--;
			}

			if (sortedDataList[lowIndex] + sortedDataList[highIndex] == sumValue) {
				// handle duplicate values
				int lowNext = lowIndex + 1;
				while ((lowNext < highIndex)
						&& (sortedDataList[lowNext] == sortedDataList[lowIndex])) {
					lowNext++;
				}
				
				int highNext = highIndex - 1;
				while ((highNext > lowIndex)
						&& (sortedDataList[highNext] == sortedDataList[highIndex])) {
					highNext--;
				}
				
				// output a list for each duplicate
				if (sortedDataList[lowIndex] == sortedDataList[highIndex]) {
					// high and low are the hold the same value
					if (lowNext == highIndex) {
						// was blocked by highIndex before end of smae value set
						lowNext++;
					}
					
					for (int low = lowIndex; low < lowNext - 1; low++) {
						for (int lowN = low + 1; lowN < lowNext; lowN++) {
							System.out.printf("pair %d(index %d) + %d(index %d) = sum value of %d\n", sortedDataList[low], low, sortedDataList[lowN], lowN, sumValue);							
						}
					}
				}
				else {
					for (int low = lowIndex; low < lowNext; low++) {
						for (int high = highIndex; high > highNext; high--) {
							System.out.printf("pair %d(index %d) + %d(index %d) = sum value of %d\n", sortedDataList[low], low, sortedDataList[high], high, sumValue);
						}
					}
				}
				
				lowIndex = lowNext;
				highIndex = highNext;
			}
		}	    	
	}
	
	
	/**
	 * entry point
	 * 
	 * @param args
	 */
	public static void main(String[] args)
	{
		SumPairList sumPairList = new SumPairList();
		
		System.out.println("First Set (values from initial problem):");
		int[] dataList1 = {1, 2, 3, 4, 5, 6, 8, 9, 12};
		sumPairList.findSumPairs(dataList1, 11);
		System.out.println();

		System.out.println("Second Set (duplicate values):");
		int[] dataList2 = {-5, 1, 3, 4, 6, 6, 6, 9, 11, 12, 14, 14, 14, 16, 19, 25};
		sumPairList.findSumPairs(dataList2, 20);
		System.out.println();

		System.out.println("Third Set (duplicate values that match with themselves):");
		int[] dataList3 = {0, 2, 4, 6, 7, 9, 11, 11, 11, 11, 15, 17, 20, 21, 22};
		sumPairList.findSumPairs(dataList3, 22);
	}
}
